from django.db.models import Avg, Count
from django.utils.datetime_safe import datetime
from knox.models import AuthToken
from knox.settings import CONSTANTS
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.permissions import AllowAny, IsAdminUser, IsAuthenticated
from rest_framework.request import ForcedAuthentication
from rest_framework.response import Response
from rest_framework.viewsets import ViewSet

from jw_nx.auth import JSONWebTokenKnoxAuthentication
from jw_nx.serializers import LoginSerializer
from jw_nx.settings import api_settings
from jw_nx.tokens import RefreshToken

response_payload_handler = api_settings.JW_NX_RESPONSE_PAYLOAD_HANDLER


class PerViewAuthenticatorMixin(object):
    authentication_classes = (JSONWebTokenKnoxAuthentication,)

    @staticmethod
    def get_authenticators_for_view(view_name):
        raise NotImplementedError("Must implement `get_authenticators_for_view` method for `JWTKnoxAPIViewSet` subclasses")

    def initialize_request(self, request, *args, **kwargs):
        """
        Returns the initial request object.
        """
        request = super(PerViewAuthenticatorMixin, self).initialize_request(request, *args, **kwargs)
        if not any([isinstance(auth, ForcedAuthentication) for auth in request.authenticators]):
            request.authenticators = self.get_authenticators()
        return request

    def get_authenticators(self):
        """
        First tries to get the specific authenticators for a view by
        calling `.get_authenticators_for_view`, but falls back on the
        class's authenticators.
        """
        authenticators = self.authentication_classes or ()

        if hasattr(self, 'action'):
            # action gets populated on the second time we are called
            per_view = self.get_authenticators_for_view(self.action)
            if per_view is not None:
                authenticators = per_view

        return [auth() for auth in authenticators]


class JWTKnoxAPIViewSet(PerViewAuthenticatorMixin, ViewSet):
    """This API endpoint set enables authentication via **JSON Web Tokens** (JWT).

    The provided JWTs are meant for a single device only and to be
    kept secret. The tokens may be set to expire after a certain time
    (see `get_token`). The tokens may be revoked in the server via the
    diverse logout endpoints. The JWTs are database-backed and may be
    revoked at any time.
    """
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def get_authenticators_for_view(view_name):
        if view_name in ('login', 'refresh'):
            return api_settings.JW_NX_LOGIN_AUTHENTICATION_CLASSES

    def get_permissions(self):
        if self.action in ('login', 'refresh'):
            return [AllowAny()]
        return super(JWTKnoxAPIViewSet, self).get_permissions()

    def get_authenticators(self):
        """
        First tries to get the specific authenticators for a view by
        calling `.get_authenticators_for_view`, but falls back on the
        class's authenticators.
        """
        authenticators = self.authentication_classes or ()

        if hasattr(self, 'action'):
            # action gets populated on the second time we are called
            per_view = self.get_authenticators_for_view(self.action)
            if per_view is not None:
                authenticators = per_view
        return [auth() for auth in authenticators]

    # @action(methods=('get',), detail=False)
    # def debug_verify(self, request):
    #     """
    #     This view returns internal data on the token, the user and the current
    #     request.
    #
    #     **NOT TO BE USED IN PRODUCTION.**
    #     """
    #     token = request.auth[0]
    #     return Response(
    #         response_payload_handler(token, request.user, request),
    #         status=status.HTTP_200_OK)

    @action(methods=['post', ], detail=False)
    def login(self, request):
        """
        This view authenticate user via USERNAME_FIELD and password
        If credential is valid => Return token
        """
        serializer = LoginSerializer(data=request.data, context={'request': request})
        serializer.is_valid(raise_exception=True)

        return Response(serializer.validated_data)

    @action(methods=('get', 'post'), detail=False)
    def verify(self, request):
        """
        This view allows a third party to verify a web token.
            if 401_UNAUTHORIZED => Token is not valid
            if 204_NO_CONTENT   => Token is valid
        """
        return Response(None, status=status.HTTP_204_NO_CONTENT)

    @action(methods=('post',), detail=False)
    def refresh(self, request):
        """ Get and validate refresh_token from request and send new access token to user """
        token = request.data['refresh_token']
        refresh = RefreshToken()
        refresh.validate_token(token)
        access = refresh.access_token
        return Response({"access_token": str(access), "refresh_token": str(refresh)})

    @action(methods=('post',), detail=False)
    def logout(self, request):
        """
        Invalidates the current token, so that it cannot be used anymore
        for authentication.
        """
        AuthToken.objects.filter(token_key=request.auth).delete()
        return Response(None, status=status.HTTP_204_NO_CONTENT)

    @action(methods=('post',), detail=False)
    def logout_other(self, request):
        """
        Invalidates all the tokens except the current one, so that all other
        remaining open sessions get closed and only the current one is still
        open.
        """
        deleted_tokens = AuthToken.objects \
            .filter(user_id=request.user.id) \
            .exclude(token_key=request.auth[:CONSTANTS.TOKEN_KEY_LENGTH]) \
            .delete()

        return Response({"Deleted tokens": deleted_tokens[0]})

    @action(methods=('post',), detail=False)
    def logout_all(self, request):
        """
        Invalidates all currently valid tokens for the user, including the
        current session. This endpoint invalidates the current token, and you
        will need to authenticate again.
        """
        deleted_tokens = AuthToken.objects.filter(user_id=request.user.id).delete()
        return Response({"Deleted tokens": deleted_tokens[0]})


class AdminAPIViewSet(PerViewAuthenticatorMixin, ViewSet):
    permission_classes = [IsAdminUser]

    @staticmethod
    def get_authenticators_for_view(view_name):
        return None

    @action(methods=('post',), detail=False)
    def delete_expired_tokens(self, request):
        del_count = AuthToken.objects \
            .filter(expiry__lt=datetime.now()) \
            .delete()[0]
        return Response({'Count': del_count}, status.HTTP_204_NO_CONTENT)

    @action(methods=('get',), detail=False)
    def average_all_per_user(self, request):
        """ Get average knox token per user """
        subquery = AuthToken.objects \
            .values('user') \
            .annotate(count=Count('token_key')) \
            .aggregate(avg=Avg('count'))
        return Response({'Average': subquery['avg']})

    @action(methods=('get',), detail=False)
    def average_active_per_user(self, request):
        average = AuthToken.objects \
            .filter(expiry__gt=datetime.now()) \
            .exclude(user_id=self.request.user.id) \
            .values('user') \
            .annotate(count=Count('pk')) \
            .aggregate(avg=Avg('count'))
        return Response({'Average active': average['avg']})

    @action(methods=('get',), detail=False)
    def average_expired_per_user(self, request):
        average = AuthToken.objects \
            .filter(expiry__lt=datetime.now()) \
            .values('user') \
            .annotate(count=Count('pk')) \
            .aggregate(avg=Avg('count'))
        return Response({'Average expired': average['avg']})



from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.models import update_last_login
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers, exceptions

from jw_nx.settings import api_settings
from jw_nx.tokens import RefreshToken
from jw_nx.utils import get_username_field

User = get_user_model()
USERNAME_FIELD = get_username_field()


class PasswordField(serializers.CharField):
    def __init__(self, **kwargs):
        kwargs.setdefault('style', {})

        kwargs['style']['input_type'] = 'password'
        kwargs['write_only'] = True

        super().__init__(**kwargs)


class LoginSerializer(serializers.Serializer):
    default_error_messages = {'no_active_account': _('No active account found with the given credentials')}

    def __init__(self, *args, **kwargs):
        super(LoginSerializer, self).__init__(*args, **kwargs)

        self.fields[USERNAME_FIELD] = serializers.CharField()
        self.fields['password'] = PasswordField()

    @staticmethod
    def validate_user(user):
        if user is None:
            msg = _('Unable to log in with provided credentials.')
            raise exceptions.ValidationError(msg)
        if not user.is_active:
            msg = _('User account is disabled.')
            raise exceptions.ValidationError(msg)

    def validate(self, attrs):
        auth_payload = {
            'request': self.context['request'],
            USERNAME_FIELD: attrs[USERNAME_FIELD],
            'password': attrs['password']
        }

        user = authenticate(**auth_payload)
        self.validate_user(user)
        refresh, access = RefreshToken().create_r_a(user)
        data = {'refresh_token': str(refresh), 'access_token': str(access)}

        if api_settings.JW_NX_UPDATE_LAST_LOGIN:
            update_last_login(None, user)

        if getattr(api_settings, 'JW_NX_RETURN_EXPIRATION', False):
            data.update({
                'access_token_expiration': access.get('exp'),
                'refresh_token_expiration': refresh.get('exp')
            })

        return data



from django.urls import include, path
from rest_framework import routers

from jw_nx.views import JWTKnoxAPIViewSet, AdminAPIViewSet

router = routers.SimpleRouter(trailing_slash=False)
router.register(r'', JWTKnoxAPIViewSet, basename='jw-nx')
router.register(r'', AdminAPIViewSet, basename='admin')

urlpatterns = [
    path('', include(router.urls)),
]



import jwt
from django.contrib.auth import get_user_model
from django.utils.translation import gettext as _
from knox.settings import CONSTANTS
from rest_framework import exceptions
from rest_framework.authentication import BaseAuthentication, get_authorization_header

from jw_nx.settings import api_settings
from jw_nx.tokens import AccessToken

User = get_user_model()
jwt_decode_handler = api_settings.JW_NX_DECODE_HANDLER
USER_ID_FIELD = api_settings.JW_NX_USER_ID_FIELD


class JSONWebTokenKnoxAuthentication(BaseAuthentication):
    """
    Clients should authenticate by passing the JWT token in the "Authorization"
    HTTP header, prepended with the `Bearer`. For example:

      Authorization: Bearer abc.def.ghi
    """
    www_authenticate_realm = 'api'

    @staticmethod
    def validate_auth(auth):
        if auth[0] != b'Bearer':
            msg = _('Invalid Authorization header. No credentials provided.')
            raise exceptions.AuthenticationFailed(msg)

        if len(auth) != 2:
            if len(auth) == 1:
                msg = _('Invalid Authorization header')
                raise exceptions.AuthenticationFailed(msg)
            elif len(auth) > 2:
                msg = _('Invalid Authorization header. Credentials string should contain no spaces.')
                raise exceptions.AuthenticationFailed(msg)
        return auth[0], auth[1]

    @staticmethod
    def validate_exp(payload):
        exp = payload.get('exp')
        if not exp:
            msg = _("Access token doesn't have expiration time.")
            raise exceptions.AuthenticationFailed(msg)
        if not isinstance(exp, int):
            msg = _("Invalid expiration type.")
            raise exceptions.AuthenticationFailed(msg)

    @staticmethod
    def get_payload(token):
        try:
            payload = jwt_decode_handler(token)
        except jwt.ExpiredSignatureError:
            msg = _('Signature has expired.')
            raise exceptions.AuthenticationFailed(msg)
        except jwt.DecodeError:
            msg = _('Error decoding signature.')
            raise exceptions.AuthenticationFailed(msg)
        except jwt.InvalidTokenError:
            msg = _('Invalid token error')
            raise exceptions.AuthenticationFailed(msg)
        except Exception as e:
            msg = _(f'Unknown error. Detail: {e}')
            raise exceptions.AuthenticationFailed(msg)

        if payload is None:
            msg = _('Error getting payload.')
            raise exceptions.AuthenticationFailed(msg)
        return payload

    def authenticate_credentials(self, payload):
        """ Returns an active user that matches the payload's user id and token. """
        user_id = payload.get(f'user_id')

        if user_id is None:
            msg = _('Invalid payload.')
            raise exceptions.AuthenticationFailed(msg)

        try:
            user = User.objects.get(id=user_id)
        except User.DoesNotExist:
            msg = _('Invalid signature.')
            raise exceptions.AuthenticationFailed(msg)

        if not user.is_active:
            msg = _('User inactive or deleted.')
            raise exceptions.AuthenticationFailed(msg)

        return user

    def authenticate(self, request):
        auth = get_authorization_header(request).split()  # [PREFIX, KEY]
        prefix, token = self.validate_auth(auth)
        user = self.authenticate_credentials(self.get_payload(token))
        access = AccessToken()
        access.validate_token(token, user)

        return user, access.payload['jkt'][:CONSTANTS.TOKEN_KEY_LENGTH]

    def authenticate_header(self, request):
        """
        Return a string to be used as the value of WWW-Authenticate
        header in a 401 Unauthorized response, or None if the
        authentication scheme should return 403 Permission Denied response.
        """

        return f'Bearer realm="{self.www_authenticate_realm}"'






from datetime import timedelta

from django.conf import settings
from rest_framework.settings import APISettings

USER_SETTINGS = getattr(settings, 'JW_NX', None)

"""
JW_NX = {
    'JW_NX_LEEWAY' : 0,
    'JW_NX_ISSUER': None,
    'JW_NX_AUDIENCE': None, 
    'JW_NX_USER_ID_FIELD': 'id',
    'JW_NX_RETURN_EXPIRATION': False,
    'JW_NX_UPDATE_LAST_LOGIN': False,
    'JW_NX_EXPIRE': timedelta(hours=10),
    'JW_NX_AUTH_HEADER_PREFIX': 'Bearer',
    'JW_NX_REFRESH_TOKEN_LIFETIME': timedelta(days=1),
    'JW_NX_ACCESS_TOKEN_LIFETIME': timedelta(minutes=5),
    'JW_NX_LOGIN_AUTHENTICATION_CLASSES' : settings.REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'],
}
"""

DEFAULTS = {
    'JW_NX_LEEWAY': 0,
    'JW_NX_ISSUER': None,
    'JW_NX_JWK_URL': None,
    'JW_NX_AUDIENCE': None,
    'JW_NX_SIGNING_KEY': settings.SECRET_KEY,
    'JW_NX_ALGORITHM': 'HS256',
    'JW_NX_USER_ID_FIELD': 'id',
    'JW_NX_VERIFYING_KEY': None,
    'JW_NX_RETURN_EXPIRATION': False,  # Optional
    'JW_NX_UPDATE_LAST_LOGIN': False,  # Optional
    'JW_NX_EXPIRE': timedelta(hours=10),
    'JW_NX_SECRET_KEY': settings.SECRET_KEY,
    'JW_NX_REFRESH_TOKEN_LIFETIME': timedelta(days=1),
    'JW_NX_ACCESS_TOKEN_LIFETIME': timedelta(minutes=5),
    'JW_NX_ENCODE_HANDLER': 'jw_nx.utils.jwt_encode_handler',
    'JW_NX_DECODE_HANDLER': 'jw_nx.utils.jwt_decode_handler',
    'JW_NX_PAYLOAD_HANDLER': 'jw_nx.utils.jwt_payload_handler',
    'JW_NX_RESPONSE_PAYLOAD_HANDLER': 'jw_nx.utils.jwt_response_payload_handler',
    'JW_NX_LOGIN_AUTHENTICATION_CLASSES': settings.REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'],

}

IMPORT_STRINGS = (
    'JW_NX_LOGIN_AUTHENTICATION_CLASSES',
    'JW_NX_ENCODE_HANDLER',
    'JW_NX_DECODE_HANDLER',
    'JW_NX_PAYLOAD_HANDLER',
    'JW_NX_PAYLOAD_GET_USERNAME_HANDLER',
    'JW_NX_PAYLOAD_GET_TOKEN_HANDLER',
    'JW_NX_RESPONSE_PAYLOAD_HANDLER',
)

api_settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)
